
#include <assert.h>
#include <string.h>
#include "Rdb.h"

#ifdef DO_INGRES
#include "RdbIngresSystem.h"
#endif
#ifdef DO_SYBASE
#include "RdbSybaseSystem.h"
#endif
#ifdef DO_INFORMIX
#include "RdbInformixSystem.h"
#endif


//---------------------------------------------------------------------------
// Data:

// List of system names that are available.
// This list must be in the order of SystemType!
const char* RdbManager::systemName[SYSTEM_COUNT] = {
    "ingres",
    "sybase",
    "informix"
  };


//---------------------------------------------------------------------------

RdbManager::RdbManager()
{

    // initialize all defined systems, be sure there is one of these
    // for each system in SystemType!

#ifdef DO_INGRES
    system[INGRES] = new RdbIngresSystem();
#else
    system[INGRES] = NULL;
#endif

#ifdef DO_SYBASE
    system[SYBASE] = new RdbSybaseSystem();
#else
    system[SYBASE] = NULL;
#endif

#ifdef DO_INFORMIX
    system[INFORMIX] = new RdbInformixSystem();
#else
    system[INFORMIX] = NULL;
#endif

}    
	
//---------------------------------------------------------------------------

RdbManager::~RdbManager()
{
    for( int i = 0 ; i < SYSTEM_COUNT ; i++ ) {
	delete system[i];
	system[i] = NULL;	// yeah yeah, not really needed...
    }
}

//---------------------------------------------------------------------------

RdbSystem* RdbManager::LookupSystem(const String _systemName) const
{
    for( int i = 0 ; i < SYSTEM_COUNT ; i++ ) {
	// typecast needed because sun's don't have strcasecmp in header
	if( strcasecmp((const char*)_systemName, systemName[i]) == 0 )
	  return LookupSystem( SystemType(i) );
    }
    return NULL;
}

//---------------------------------------------------------------------------

RdbSystem* RdbManager::LookupSystem(const RdbManager::SystemType _systemType)
     const
{
    assert( /* _systemType >= 0 && */  _systemType < SYSTEM_COUNT);
    return system[_systemType];
}


//---------------------------------------------------------------------------

RdbError::ErrorCode RdbManager::Commit()
{
    // update this to use 2 phase commit
    for( int i = 0 ; i < SYSTEM_COUNT ; i++ ) {
	if( system[i] != NULL ) {
	    system[i]->CommitAllDatabases();
	}
	// this will abort on error later... 
    }
    return NO_ERROR;		// this will be better later...
}


//---------------------------------------------------------------------------

RdbError::ErrorCode RdbManager::Rollback()
{
    for( int i = 0 ; i < SYSTEM_COUNT ; i++ ) {
	if( system[i] ) {
	    system[i]->RollbackAllDatabases();
        }
    }
    return NO_ERROR;		// this will be better later...
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
