#ifndef _RDBINGRESDATABASE_H_
#define _RDBINGRESDATABASE_H_

#include "RdbDatabase.h"
#include "String.h"

// Forward declarations
class RdbIngresSystem;



class RdbIngresDatabase
: public RdbDatabase
{
  public:
    
    RdbIngresDatabase(RdbIngresSystem& ingresSystem,
		      const String newDbName,
		      const String userid = String(),
		      const String password = String(),
		      const String serverName = String());
				// open database called newDbName

    ~RdbIngresDatabase();

    virtual RdbQuery* NewQuery(RdbJoin& joinInfo, const BitVector& adorn);
				// allocate a query on the joinInfo

    virtual RdbInsert* NewInsert(RdbRelation& rel);
				// allocate an insert statment on rel

    virtual RdbDelete* NewDelete(RdbRelation& rel, const BitVector& adorn);
				// allocate a delete statment on rel

    RdbIngresSystem& System() const; // return the system for the database

    virtual void const* Implementor() const;
				// virtual access to 'this'

  private:

    static int nextSessionId;	// next available session id
				// when this rolls over, all other sessions
				// had better be closed!!

    int sessionId;		// the identifier of this db to the backend

    RdbIngresSystem& system;

    // see RdbDatabase:

    virtual ErrorCode VConnect();

    virtual ErrorCode VDisconnect();

    virtual ErrorCode VImmediateSql(const String& sql);

    virtual ErrorCode VCommit();

    virtual ErrorCode VRollback();

    virtual ErrorCode VPrepare(const String& id, const String& statement);

    virtual String VGetErrorMessage();

    virtual ErrorCode VSetFocus();

    virtual ErrorCode VRemoveFocus();

    RdbIngresDatabase(const RdbIngresDatabase&); // na

    RdbIngresDatabase& operator =(const RdbIngresDatabase&); // na

};


#include "RdbIngresDatabase.i"

#endif // _RDBINGRESDATABASE_H_
