#ifndef _LIST_H_
#define _LIST_H_

// The classes below represent a dynamic ordered list (array).

#include <stdlib.h>
#include <assert.h>
#include <iostream.h>


// A dynamic arry of void pointers.  It does not copy or delete items,
// it just stores the pointers.

class PtrListBase
{
  public:
    
    PtrListBase(int startSize = 10); // creates a list of startSize elements

    PtrListBase(const PtrListBase& other); // copy

    ~PtrListBase();		// deletes the list, and all of the elements

    PtrListBase& operator =(const PtrListBase& other); // copy

    int Size() const;		// returns the number of elements in the list

    int Append(void* element, int n = 1);
				// inserts N copies of element at the end
				// of the list and returns the index of 
				// where the first element was stored

    int Insert(void* element, int loc);
				// inserts an element into the list at location
				// loc.  loc must be >= 0 & <= Size()
				// returns loc

    void* Delete(int i);	// deletes element i from the list
				// any index values are then invalidated

    void DeleteAll();		// deletes all of the enteries
    
    int ValidSubscript(int i) const; // determines if a subscript 
				     // is valid or not

    void* operator [](int i) const; // pointer to item i;

    void* Set(int i, void* element); // set's item i to element
				     // i must be a valid item

  private:

    void Resize(int s);		// set the size of the array to at least s

    void** list;		// array of the object ptrs
    int count;			// last used element in the array
    int allocated;		// size of the allocated array
};


// A dynamic array of class T pointers.  No constructing or destructing.

template <class T>
class PtrList
: private PtrListBase
{

  public:

    PtrList(int startSize = 10); // make room for startSize elements

    PtrList(const PtrList<T>& other);

    ~PtrList();

    PtrList& operator =(const PtrList& other);

    T* operator[](int i) const; // reference to item i

    int Append(T* element, int n = 1);
				// insert n copies of item at end,
				// return subscript of first

    int Insert(T* element, int loc);
				// inserts an element into the list at location
				// loc.  loc must be >= 0 & <= Size()
				// returns loc

    T* Set(int i, T* element);	// set's item i to element
				// i must be a valid item

    T* Delete(int i);		// delete item i from the list,
				// return a ptr to that item

    PtrListBase::Size;
    PtrListBase::DeleteAll;
    PtrListBase::ValidSubscript;
};


// This list allows you to insert pointers to objects,
// and it will delete the object on a delete from list (or destruction of list)

template <class T>
class DList
: private PtrList<T>
{

  public:

    DList(int startSize = 10); // make room for startSize elements
    
    DList(const DList<T>& other); // copy

    ~DList();			// destructs all elements

    DList<T>& operator =(const DList<T>& other); // copy

    void Delete(int i);		// destruct object i, shift all subscripts

    void DeleteAll();		// destruct all objects

    void Set(int i, T* element); // set item i to element, destory old element
				 // i must be a valid subscript

    PtrList<T>::operator [];
    PtrList<T>::Append;
    PtrList<T>::Insert;
    PtrList<T>::Size;
    PtrList<T>::ValidSubscript;
};


// this list makes a separate copy of the items, and destruct them

template <class T>
class List
: private DList<T>
{
  public:

    List( int startSize = 10 );	// make room for startSize elements

    List( const List<T>& other ); // copy

    ~List();

    List<T>& operator =( const List<T>& other ); // copy

    int Append( const T& element ); // add a copy of element to the list

    int Insert(const T& element, int loc);
				// inserts an element into the list at location
				// loc.  loc must be >= 0 & <= Size()
				// returns loc

    void Set(int i, const T& element); // add copy of element to list at i,
				       // delete the old item. i must be valid

    T& operator [](int i) const;

    DList<T>::Size;
    DList<T>::ValidSubscript;
    DList<T>::Delete;
    DList<T>::DeleteAll;
};  


// performs insertion sort to keep this list in order

template <class T>
class SortedList
: private List<T>
{
  public:

    SortedList( int startSize = 10 ); // make room for startSize elements

    SortedList( const SortedList<T>& other ); // copy

    ~SortedList();

    SortedList<T>& operator =(const SortedList<T>& other); // copy

    int Insert( const T& element ); // add a copy of element to the list
				    // in sorted order

    List<T>::Size;
    List<T>::ValidSubscript;
    List<T>::Delete;
    List<T>::DeleteAll;
    List<T>::operator [];
};  




#include "List.i"

#endif // _LIST_H_
