#include <stdlib.h>
#include <sys/unistd.h>
#include <sys/time.h>
#include <strstream.h>
#include <iomanip.h>
#include "sybase.h"
#include "RdbSybaseSystem.h"
#include "RdbSybaseDatabase.h"



// static prototypes:
static int ErrorHandler(DBPROCESS* dbproc, int severity, int dberr, int oserr,
			char* dberrstr, char* oserrstr);
static int MessageHandler(DBPROCESS* dbproc, DBINT msgno, int msgstate, 
			  int severity, char* msgtext, char* srvname,
	//  char* procname, unsigned short line);
    		  char* procname, int line);
	//  char* procname, DBUSMALLINT line);


RdbSybaseSystem::RdbSybaseSystem()
: RdbSystem()
{
    // oid stuff:
    pid = (pid_t) getpid();
    time_ms = time(NULL);
    serialNum = 0;

    // db-lib initialization:
    if( dbinit() == FAIL )	// initialize dblib
      Error(SQL_ERROR);
    dberrhandle(ErrorHandler);	// establish callback routines
    dbmsghandle(MessageHandler);
}


RdbSybaseSystem::~RdbSybaseSystem()
{
    dbexit();
}


RdbDatabase* RdbSybaseSystem::OpenDatabase(const String dbName,
					   const String userid,
					   const String password,
					   const String serverName)
{
    return(new RdbSybaseDatabase(*this,dbName, userid, password, serverName)); 
}


// Sybase call back when errors are detected
static int ErrorHandler(DBPROCESS* dbproc, int severity, int dberr, int oserr,
			char* dberrstr, char* oserrstr)
{
    // this is the default handler, add to this to make more robust!
    if( dbproc == NULL || DBDEAD(dbproc) ) {
	cerr << "using a dead dbproc, aborting!\n";
	return INT_EXIT;
    }
    cerr << "DB-Library error:\n\t" << dberrstr << '\n';
    if( oserr != DBNOERR )
      cerr << "Operating system error:\n\t" << oserrstr << '\n';
    return INT_CANCEL;
}


// Sybase call back when messages are reported
static int MessageHandler(DBPROCESS* dbproc, DBINT msgno, int msgstate, 
			  int severity, char* msgtext, char* srvname,
		//	  char* procname, unsigned short line)
			  char* procname, int line)
			  //char* procname, DBUSMALLINT line)
{
    // this is the default handler, add to this to make more robust!
    // see dblib ref pg. 2-180 from more info
#if defined(DEBUG)
    // note: this should be used, with message filtering or something
    // messages could go to a file...
    cerr << "\nDB-Library message: "
      << "Msg " << msgno << ", level " << severity << ", state " << msgstate;
    if( *srvname )
      cerr << ", Server " << srvname;
    if( *procname )
      cerr << ", Procedure " << procname;
    if( line )
      cerr << ", Line " << line;
    cerr << "\n\t" << msgtext << '\n';
#endif
    return 0;
}


String RdbSybaseSystem::GetOid()
{
    ostrstream s;
    s.fill('0');
    s << "rdb" << hex << setw(8) << pid << setw(8) << time_ms 
      << setw(8) << serialNum++;
    return String(s);
}
