// *************************************************************************
//    BuiltinRelation.C
// *************************************************************************

#include "BuiltinRelation.h"

#include "Config.h"
#include "Globals.h"
#include "CoralError.h"
#include "InitBuiltins.h"
#include "Interface.h"        // add builtin
#include "StackMark.h"
#include "TupleIterator.h"
#include "CArg.h"
#include "Debug.h"

// *************************************************************************
int Cor_BuiltinRelation::insert_tuple(Cor_Tuple *tuple)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_ArgList& args = tuple->args();

    int env_size = args.max_vars();

#ifdef    DEBUG
    if (env_size)
        cerr << "Warning: attempting to insert a non-ground fact in "
          << __FUNCTION__ << endl;
#endif

    Cor_StackMark stackmark0;

    Cor_BindEnv *env;

    if (env_size)
        env = new Cor_BindEnv (env_size);
    else
        env = NULL;

    Cor_TupleIterator it(this, args, env);

    it.get_next();
    stackmark0.pop_to();

    delete env;
    return(0);
}

// *************************************************************************
int Cor_BuiltinRelation::insert_new(Cor_ArgList &args, Cor_BindEnv *env)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_BindEnv *new_env = NULL;;
    int env_size;

    if (env == NULL) {
        env_size = args.max_vars();
#ifdef    DEBUG
        if (env_size) {
            cerr << "Warning: attempting to insert a non-ground fact in "
              << __FUNCTION__ << endl;

            cerr << "The tuple is: " << args << endl;
        }
#endif
        if (env_size)
          new_env = new Cor_BindEnv(env_size);
        else
          new_env = NULL;
    } else
        new_env = env;

    Cor_StackMark stackmark0;

    Cor_TupleIterator it(this, args, new_env);
    it.get_next();
    int stat = !(it.no_match());
    stackmark0.pop_to();

    if (new_env && (new_env != env))
        delete new_env;

    return(stat);
}

// *************************************************************************
Cor_BuiltinRelation::Cor_BuiltinRelation(int a, Cor_CArg *n,
    SolveProc solve_proc, int insert): Cor_Relation()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    count = -1;  // NOTE: this is a kludge to easily identify a builtin relation
    single_success = 1;
    _arity = a;
    name = *n;
    solver = solve_proc;
    ref_count = 1;
    if (insert)
        Cor_InsertBuiltin(name, this);
}

// *************************************************************************
Cor_BuiltinRelation::Cor_BuiltinRelation(char *predname, int arity,
        UserRelProc solver) : Cor_Relation()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    add_builtin(predname, arity, solver);
}

// *************************************************************************
Cor_BuiltinRelation::Cor_BuiltinRelation(char *predname, int arity,
    UserTupleProc solver) : Cor_Relation()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    add_builtin(predname, arity, solver);
}

// *************************************************************************
void Cor_BuiltinRelation::print_name(ostream *out)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (is_printable_builtin()) {
        if (arity() >= 0) {
            *out << name.symbol() << "/" << arity()
                 << "\t : (builtin)\n";
        } else {
            *out << name.symbol() << "/-\t : (builtin)\n";
        }
    }
}

// *************************************************************************
Cor_BindEnv *Cor_BuiltinRelation::get_next(Cor_TupleIterator &iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (iterator.ipos > 0) {
        iterator.stack_mark.pop_to();
        iterator.set_no_match();
#        ifdef DO_TRACE
            if (Cor_exEnv->dbg_get_next) {
                *(Cor_exEnv->trace_strm)
                    << "Builtin Relation: " << name.symbol()
                    << " - no more answers\n";
            }
#        endif
        return(iterator.bindenv);
    }

#    ifdef DO_TRACE
    if (Cor_exEnv->dbg_get_next) {
        *(Cor_exEnv->trace_strm) << 
            "Builtin Relation: Starting on " <<  name.symbol() << " ( ";
        iterator.arg_list.print(iterator.bindenv, Cor_exEnv->trace_strm);
        *(Cor_exEnv->trace_strm) << ")\n";
    }
#    endif

    iterator.stack_mark.get_mark();

    iterator.ipos++;
    return(solver(*this, iterator));
}

// *************************************************************************
void Cor_BuiltinRelation::release(Cor_TupleIterator &)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // Do nothing.  Subclasses of builtin relation that store structures in
    // TupleIterator may redefine this function to release allocated space.
}

// *************************************************************************
