
//**************************************************************************
//    Connection.h
//
//    The Connection Class - Used to hold all information about a 
//    connection.
//**************************************************************************

#ifndef _Connection_H_
#define _Connection_H_

#include "User.h"
#include "ParseStack.h"

//**************************************************************************

                      // The number of logical users to 
                      // allow on a single physical connection
                      // (Actually, one less "real" user
                      // that this number is allowed to
                      // give room for a default user "nobody"
#define  MAX_USRS_PER_CONN         17

                      // A special user that always exists if there is
                      // a connection.  This is needed so that an exEnv
                      // exists so that you can enter commands (like login()).
                      // This value should not be changed, it is here only
                      // to clarify the code.
#define  NOBODY    0

                      // Max length of string in cwd variable
#define  MAX_PATH_LEN  256
//**************************************************************************

class Cor_Connection {

friend class Cor_ConnectArray;

protected:

  int         num_users;    // A count of the number of logical users

  int         nxt_user;     // A spot "likely" to be free in the userArray

  int         cur_user;     // The index of the current user

public:  

  int         conn_id;      // The index of this connection

                            // An array of users
  Cor_User       *userArray[MAX_USRS_PER_CONN];

  Cor_ParseStack  pStack;       // One ParseStack needed per connection

  CoralStream *conn_strm;       // The stream for this connection

  int          sock_fd;         // So we don't lose track of the socket
                                // if it closes weirdly.

  char         cwd[MAX_PATH_LEN];  // Where this connection is at.

  Cor_Connection(CoralStream *io_strm);
 ~Cor_Connection();

    // Switch to this connection
  int useConnection();

    // Switch to this user
  int useUser(int user_id);

    // Generates a new user
  int  NewUser(char *name, char *passwd);

    // Deletes a user from the array of users
  void DeleteUser(int user_id);

    // Find the user with this name, returning the first match
  int FindUser(char *name);

    // Return a pointer to the current user
  inline Cor_User *currentUser() {
      return(userArray[cur_user]);
  }

    // Returns a pointer to the user specified (by user_id) 
  inline Cor_User *indxUser(int user_id) {
      return(userArray[user_id]);
  }

    // List all users for this connection, return the number found
  int ListWho(CoralStream *strm);
};

//**************************************************************************

#endif  // _Connection_H_
