
//**************************************************************************
//    LocalConnectArray.C
//
//    The LocalConnectArray Class - Used to implement a single console
//    (local) connnection, but still be derived from the abstract
//    ConnectArray class.
//
//    A ConnectArray holds an array of Connections, of which each
//    holds an array of Users.
//     
//**************************************************************************

#include <string.h>

#include "LocalConnectArray.h"
#include "Globals.h"
#include "ClientProtocol.h"
#include "Debug.h"

//**************************************************************************
// LocalConnectArray Constructor - Sets up a console connection.
//**************************************************************************
Cor_LocalConnectArray::Cor_LocalConnectArray()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    cur_conn = 0;
    max_conn = 1;

#ifdef    DO_EXPLAIN
    expl_serv_strm = NULL;
#endif

      // Make a local connection
    connArray[0] = new Cor_Connection(NULL);
    connArray[0]->conn_id = 0;

    num_conn = 1;

      // Start using the connection
    connArray[0]->useConnection();
}

//**************************************************************************
// LocalConnectArray Destructor - Closes the explain connection (if any),
//    then closes each connection in the array.
//**************************************************************************
Cor_LocalConnectArray::~Cor_LocalConnectArray()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

#ifdef    DO_EXPLAIN
      // Close explain server socket
    delete expl_serv_strm;
#endif

      // Close the connection
    delete connArray[0];
}

//**************************************************************************
// ExplainConnection() -
//    This function is callled when a connection request is recieved on the
//    explain port.  This function decides if the connection attempt is
//    valid and accepts or refuses it.
//
//    Return: The return value is zero if everything is okay
//        or < 0 if an error occurs or the connection is refused.
//
//**************************************************************************
int Cor_LocalConnectArray::ExplainConnection()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef     ADD_LATER
#ifdef    DO_EXPLAIN

    char buff[500];
    int c_id=0, u_id=0, pnt=0;

    if (!expl_serv_strm) 
        return(-1);

    if (in->gets(buff, 500)) {
        return(-1);
    }

    sscanf(buff, "%d %d %n", &c_id, &u_id, &pnt); 
   
    if (u_id){ 
        for (int i=pnt; buff[i]; i++)
            if ((buff[i] == '\r') || (buff[i] == '\n'))
                buff[i] = '\0';
    } else
        buff[pnt] = '\0';

    Cor_Connection *conn = connArray[c_id];
    Cor_User *user;

    if ((! conn) || (! (user = conn->userArray[u_id])) ||
            strcmp(user->password,&buff[pnt]) || 
            (user->Cor_exEnv.explain_strm != NULL) || 
            ! (user->Cor_exEnv.GlobalRelationOptions & REL_EXPLAIN)){

        int sd = tmp_sock->socketfd();
        char *mess = "Connection refused by Coral Server.\n";
        write(sd,mess,strlen(mess));

          // Disconnect
        shutdown(sd,2);
        close(sd);
        return(0);
    }

      // They check out, set them up to recieve explain info
    user->explain_sock = tmp_sock;
    user->Cor_exEnv.explain_strm = new CoralStream(tmp_sock);
      // And inform client (client blocks waiting for a message)
    *(user->Cor_exEnv.out_strm) << endl << CP_MESG
            << "Explain Connected.\n" << flush;
    *(user->Cor_exEnv.explain_strm) << "coral.\n";

#endif    // DO_EXPLAIN

#endif    // ADD_LATER

    return(0);
}

//**************************************************************************
// ListWho() - 
//    This function lists all users for the connection specified, 
//    (printing the output to the stream given), if the connection
//    specified is NULL (the default value), lists all users on all
//    connections.
//
//    Return: The number of users on this (or all) connections.
//
//**************************************************************************
int Cor_LocalConnectArray::ListWho(CoralStream *strm, Cor_Connection *conn)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int cnt = 0;

    if (conn){
        cnt += conn->ListWho(strm);
    } else {
        if (connArray[0])
            cnt += connArray[0]->ListWho(strm);
    }

    return(cnt);
}

//**************************************************************************
