
//**************************************************************************
//    User.C
//
//    The User Class - Contains all the information about a user and
//    their current envirornment.
//
//**************************************************************************

#include "User.h"

#include "CoralInit.h"
#include "Database.h"
#include "Debug.h"
          
//**************************************************************************
// User Constructor -
//    Clears everything, and sets all envirornment values to the defaults.
//
//**************************************************************************
Cor_User::Cor_User(CoralStream *in, CoralStream *out, CoralStream *err)
            :Cor_exEnv(Cor_deflt_exEnv,in,out,err)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
   *login_name = *password = '\0';
    user_id = 0;

      // Set current database to the default
    Cor_CArg db_arg = Cor_CArg(COR_DEFAULT_DB);
    Cor_CurDB = Cor_lookupDB(db_arg);

    num_answers  = DEFLT_ANSWRS_PER_QRY;
    qid = 0;

    collector    = NULL;
    explain_strm = NULL;
}

//**************************************************************************
// User Destructor -
//    Deletes the socket to explain, if the user has one.
//
//**************************************************************************
Cor_User::~Cor_User()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
      // Close up connection to explain (if any)
    delete explain_strm;
}

//**************************************************************************
// whoami() -
//    A simple function to print the user's name on the stream given.
//
//**************************************************************************
void Cor_User::whoami(CoralStream *strm)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *strm << login_name << endl;
}

//**************************************************************************
// yield() -
//    Set flag(s) to indicate a yield should be performed at the end of
//    this command.
//
//**************************************************************************
void Cor_User::yield()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
      // Talking about this user's exEnv, not global.

    Cor_exEnv.C_interrupt_raised = 2; // informs parser odd things are occuring
    Cor_exEnv.yield = 1;              // (in this case, yielding)
}

//**************************************************************************
