//**************************************
// GraceJoin.C     
//
// Ravi Murthy
//**************************************

#ifdef    WITH_PERSISTENCE
#include <assert.h>
#include "GraceJoin.h"

const int NUM_PARTITION = 5;
Cor_JoinAttrs* find_joinattrs(Cor_TupleIterator *outer,
            Cor_TupleIterator *inner);

// The constructor partitions the outer and the inner relation into some
// number of partitions.
Cor_GraceJoin::Cor_GraceJoin(Cor_TupleIterator* left, Cor_TupleIterator *right)
    :Cor_ExtJoin(left, right)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    // Initialize private members
    new_part_start = 1;
    next_part_num = 0;
    innerpart_iter = NULL;
    outerpart_iter = NULL;
    curr_hashjoin = NULL;
    
    // First determine the join attrs
    joinattrs = find_joinattrs(left, right);

    assert(joinattrs != NULL);
    // Partition the outer relation
    outer_part = new Cor_Partition(NUM_PARTITION);
    if (!(outer_part->FormPartitions(left, joinattrs))) {
        *(Cor_exEnv->err_strm)
            << "Error in forming partitions of the outer relation\n";
        return;
    }

    // Partition the inner relation
    inner_part = new Cor_Partition(NUM_PARTITION);
    if (!(inner_part->FormPartitions(right, joinattrs))) {
        *(Cor_exEnv->err_strm)
            << "Error in forming partitions of the inner relation\n";
        return;
    }
}

// Destructor
Cor_GraceJoin::~Cor_GraceJoin()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // Deallocate partitions

    delete outer_part;
    delete inner_part;

    // Deallocate Cor_JoinAttrs
    delete joinattrs;

}

//*****************************************

// Return the next result tuple from the grace join.
Cor_Tuple *Cor_GraceJoin::get_next_tuple(Cor_TupleIterator& it)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    while (1) {
        if (new_part_start) {       
            // Have to start with new partitions
            // First delete iterators and join for the previous partitions
            delete innerpart_iter;
            delete outerpart_iter;
            delete curr_hashjoin;
    
            if (next_part_num >= NUM_PARTITION) {
                // No more partitions ...end of join
                it.set_no_match();
                return NULL;
            }
    
            // Initialize iterators over the next pair of partitions
    
            innerpart_iter = inner_part->getIterator(next_part_num,
                        right_rel->arg_list, it.bindenv);
    
            curr_hashjoin = new Cor_HashJoin(innerpart_iter, joinattrs);
    
            outerpart_iter = outer_part->getIterator(next_part_num,
                    left_rel->arg_list, it.bindenv);
            new_part_start = 0;
            next_part_num++;
        }
    
        curr_hashjoin->get_next_tuple(outerpart_iter);
    
        if (!(outerpart_iter->no_match())) {  // Join tuple found
            it.reset_no_match();
            return NULL;
        }
    
        // No more tuple from this pair of partitions
        new_part_start = 1;
    }
}

//******************************************************
// Find the attributes to be joined between the two relations.
// Returns a Cor_JoinAttrs object
Cor_JoinAttrs* find_joinattrs(Cor_TupleIterator *outer,
        Cor_TupleIterator *inner)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int size = outer->arg_list.count();

    // First store var nums for all variables in outer
    int *attrlist = new int[size];
    Cor_CArg *var1;

    for (int i = 0; i < outer->arg_list.count(); i++) {
        var1 = &(outer->arg_list[i]);
        attrlist[i] = var1->var_num();
    }


    // We will store var nums of all common variables in this array
    int *common = new int[size];
    int common_count = 0;

    // For each variable in inner, check if it has occurred in the outer
    for (i = 0; i < inner->arg_list.count(); i++) {
        var1 = &(inner->arg_list[i]);
        int j = 0;
        while ((j < size) && (attrlist[j] != var1->var_num()))
            j++;

        // Found a common attr.
        if (j < size) {
            common[common_count] =  var1->var_num();
            common_count++;
        }        
    }

    delete attrlist;
    return (new Cor_JoinAttrs(common, common_count));
}

// *************************************************************************
#endif // WITH_PERSISTENCE
