
// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
// 
//   CallStk.C  - keep track of the call stack (for debugging)
// 
//   Shaun Flisakowski, 1996
// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

#include  <string.h>

#include  "CallStk.h"

// Only keeps track of these things when DO_CALL_STACK is set.

  // Start at this size, then double
#define DEFLT_CALLSTK_LEN    250
#define DEFLT_STRTAB_LEN      64

CallStk  *callstk = NULL;

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
CS_StrTab::CS_StrTab()
{
    num = 0;
    length = DEFLT_STRTAB_LEN;
    data = new char* [length];
    data[0] = NULL;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
CS_StrTab::~CS_StrTab()
{
    for (unsigned int j=0; j < num; j++) {
        delete [] data[j];
    }

    delete [] data;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
// returns index in the table
unsigned int CS_StrTab::add(const char *str)
{
    unsigned int j;

      // This is more likely
    if (num && (strcmp(str,data[num-1]) == 0))
        return(num-1);

    for (j=0; j < num; j++) {
        if (strcmp(str,data[j]) == 0)
            return(j);
    }

       // Expand the array?
    if (num == length) {
        char **olddata = data;

        length += length;
        data = new char* [length];

        for (j=0; j < num; j++)
            data[j] = olddata[j];

        delete [] olddata;
    }

    int len = strlen(str);

    data[num] = new char [len+1];
    strcpy(data[num],str);

    return(num++);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
// get the string pointed to by index (stays in table)
const char *CS_StrTab::get(unsigned int index)
{
    if (index < num)
        return(data[index]);
    else
        return(NULL);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
CallStk_Mark::CallStk_Mark( char *func, char *file, int lineno )
{
    if (callstk) {
        item = new CallStk_Item(func,file,lineno);

         // Push it onto the global stack
        callstk->entry(item);
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
CallStk_Mark::~CallStk_Mark()
{
    if (callstk) {
          // Push it onto the global stack
        callstk->exit(item);
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
CallStk_Item::CallStk_Item( char *_func, char *_file, int _lineno )
{
    func = callstk->funcnames.add(_func);
    file = callstk->filenames.add(_file);
    lineno = _lineno;
    active = 1;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
CallStk_Item::~CallStk_Item()
{
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
ostream& operator <<(ostream& out, const CallStk_Item& itm)
{
    out << callstk->filenames.get(itm.file)
        << ":" << itm.lineno << ":"
        << callstk->funcnames.get(itm.func);
    return(out);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
CallStk_Entry::CallStk_Entry()
{
    type = Function_UnUsed;
    item = NULL;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
CallStk_Entry::~CallStk_Entry()
{
    if (type == Function_Entry)
        delete item;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
CallStk_Entry& CallStk_Entry::operator =(const CallStk_Entry& src)
{
    type = src.type;
    item = src.item;

    return(*this);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
ostream& operator <<(ostream& out, const CallStk_Entry& ent)
{
    switch (ent.type) {

    default:
    case Function_UnUsed:
         break; 

    case Function_Entry:
         out << "Enter:" << *(ent.item) << endl;
         break; 

    case Function_Exit:
         out << "Leave:" << *(ent.item) << endl;
         break; 
    }

    return(out);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
CallStk::CallStk()
{
    num = 0;
    len = DEFLT_CALLSTK_LEN;
    
    stk = new CallStk_Entry[len];
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
CallStk::~CallStk()
{
    delete [] stk;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void CallStk::checksize()
{
    if (num == len) {
        // Need to make stack larger (double it)
        len += len;
        CallStk_Entry *oldstk = stk;
        stk = new CallStk_Entry [len];

        for (int j=0; j < num; j++) {
            stk[j] = oldstk[j];
            oldstk[j].type = Function_UnUsed;
        }

        delete [] oldstk;
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void CallStk::entry(CallStk_Item *item)
{
    checksize();
    stk[num].type   = Function_Entry;
    stk[num++].item = item;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void CallStk::exit(CallStk_Item *item)
{
    checksize();

    item->active    = 0;
    stk[num].type   = Function_Exit;
    stk[num++].item = item;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
// Squash the stack, removing unused entries
void CallStk::compress()
{
    int j,k;

    for (j=0; j < num; j++)
        if (stk[j].type == Function_UnUsed)
            break;

      // We have a hole?
    if (j < num) {
        for (k=j+1; k < num; k++) {
            if (stk[k].type != Function_UnUsed) {
                stk[j++] = stk[k];
                stk[k].type = Function_UnUsed;
            }
        }

        num = j;
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
// Throw away the stack trace for all inactive entries
void CallStk::reset()
{
    int j;

    if (!this)
        return;

    for (j=0; j < num; j++)
        if (! stk[j].item->active) {
            if (stk[j].type == Function_Entry)
                delete stk[j].item;
            stk[j].item = NULL;
            stk[j].type = Function_UnUsed;
        }

    compress();
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
// Show the call stack, to depth given (0 == "all")
void CallStk::show_call_stack(ostream* out, int depth)
{
    if (!this)
        return;

    *out << "\nCall Stack:\n";
    for (int j=num-1; j >= 0; j--)
        if (stk[j].item->active) {
            *out << stk[j];
             if (--depth == 0)
                break;
        }
    *out << endl;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
// Show all calls in the trace
void CallStk::show_call_trace(ostream *out)
{
    if (!this)
        return;

    *out << "\nCall Trace:\n";
    for (int j=0; j < num; j++)
        *out << stk[j];
    *out << endl;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
#undef DEFLT_CALLSTK_LEN
#undef DEFLT_STRTAB_LEN
