// **************************************************************************
//    Interface.h
// **************************************************************************

#ifndef _Interface_H_
#define _Interface_H_

#include <stdarg.h>

#include "ArgList.h"
#include "BuiltinSymbols.h"
#include "BuiltinRelation.h"
#include "TupleIterator.h"
#include "ParserStruct.h"
#include "CArg.h"
#include "Stack.h"

// **************************************************************************

class Cor_HashSimpleRelation;

class Cor_Relation;
typedef Cor_Relation* (*UserRelProc)(Cor_Tuple  *query);

// **************************************************************************
//
//  These structures are used in the C++ code created by the
//  CORAL/C++ translator


class Cor_ExecutionList {

  public:
    Cor_ParserStruct   **exec_terms;

    Cor_CArgStack        params;
    int                  no_of_params;

    int                  max_no_of_terms;
    int                  actual_no_of_terms;
    int                  step_size;
    int                  initialized;

    void add(Cor_ParserStruct *pstruct);

    void initialize(int i, int num_params, void * /* not_used */) {
      exec_terms = new Cor_ParserStruct*[i];
      params.empty();
      no_of_params = num_params;
      max_no_of_terms = i;
      step_size = i;
      actual_no_of_terms = 0;
      initialized = 1;
    }

    int execute();
    int execute(int index, int& param_index);

    Cor_ExecutionList() { 
        actual_no_of_terms = 0;
        exec_terms = NULL;
    }

   ~Cor_ExecutionList() { 
        params.empty();
        for (int j=0; j < actual_no_of_terms; j++)
            delete exec_terms[j];
        delete [] exec_terms;
    }
};

// **************************************************************************
/*
 *    For embedded coral programs:
 *
 *    call init_coral() at the beginning of main.
 *    call exit_coral() at the end of main.
 */
// **************************************************************************
/* 
 * These structures are used to provide C++ users with the facility
 * of scans over CORAL relations
 */
class Cor_ScanDesc {

  private:

    Cor_TupleIterator *it;
    Cor_BindEnv       *env;
    Cor_ArgList       *arg_list;
    Cor_StackMark     *stackmark;
    int                tuple_fetched;

  public:

    Cor_ScanDesc(Cor_Relation *rel);
    Cor_ScanDesc(Cor_Relation *rel, Cor_Tuple *match_tuple);
   ~Cor_ScanDesc();
    Cor_Tuple *next_tuple();
    int no_match() {
      if (this)
          return(it->no_match());
      else
          return(1);
    }
};

// **************************************************************************
/* These extern function declarations are used by the translated C++
 * program, with the embedded CORAL code removed
 */

extern int  Cor_consultFile(char *, char *);
extern int  Cor_consultFile(istream *, char *);
extern int  Cor_consultBuff(char *, char *);
extern void Cor_processInputStack();


// **************************************************************************
/* The following functions are defined in Interface.C */

/* -------------------------------------------------------------------- */
/* Interface functions for creating arguments of tuples:        */

extern Cor_CArg *make_arg  (char *name);
extern Cor_CArg *make_arg  (Cor_CArg& name);

extern Cor_CArg *make_var  (int var_num);
extern Cor_CArg *make_var  (char *print_name, int var_num);
extern Cor_CArg *make_var  (Cor_CArg *print_name, int var_num);

extern Cor_CArg *make_arg  (char *func_name, Cor_ArgList *args);

extern Cor_CArg *make_cons (Cor_CArg *car, Cor_CArg *cdr);

extern Cor_CArg *make_nil  ();

        // These are not arg lists, they are real lists (ie [a,b,c])
        // All items in vararg list should be of type Cor_CArg*
extern Cor_CArg  *make_list (int num_arg, ...);
extern int        list_length( Cor_CArg *list );

        // All items in var_arg list should be of type Cor_CArg*
extern Cor_ArgList *make_arglist (int len, ...);
        // A destructive version of the above (deletes arguments)
extern Cor_ArgList *make_arglist_d (int len, ...);

        // Versions which take CArg references instead 
extern Cor_ArgList *make_arglist (const Cor_CArg& a1);
extern Cor_ArgList *make_arglist (const Cor_CArg& a1, const Cor_CArg& a2);
extern Cor_ArgList *make_arglist (const Cor_CArg& a1,
        const Cor_CArg& a2, const Cor_CArg& a3);

extern Cor_ArgList *concat_arglists (Cor_ArgList *al1, Cor_ArgList *al2);
extern Cor_ArgList *make_vararglist (int n);

extern Cor_ScanDesc *open_scan(char *example_fact);

/* -------------------------------------------------------------------- */
/* Interface functions for testing and getting C++ data structures from 
        CORAL data:                         */

/** int Cor_Tuple::arity ();   and 
    Cor_CArg& Cor_Tuple::arg (int arg_pos);   **/

/** int Cor_ArgList::arity ();  and 
    Cor_CArg& Cor_ArgList::arg (int arg_pos);  **/


inline int is_int    (const Cor_CArg& c) { return c.is_int(); }
inline int is_long   (const Cor_CArg& c) { return c.is_long(); }
inline int is_short  (const Cor_CArg& c) { return c.is_short(); }
inline int is_float  (const Cor_CArg& c) { return c.is_float(); }
inline int is_double (const Cor_CArg& c) { return c.is_double(); }
inline int is_string (const Cor_CArg& c) { return c.is_string(); }
inline int is_chunk  (const Cor_CArg& c) { return c.is_chunk(); }

extern int is_num      (const Cor_CArg& c);
extern int is_constant (const Cor_CArg& c);

extern int is_var      (const Cor_CArg& c);
extern int is_functor  (const Cor_CArg& c);
extern int is_list     (const Cor_CArg& c);
extern int is_nil      (const Cor_CArg& c);

extern int    make_int     (const Cor_CArg& c);
extern short  make_short   (const Cor_CArg& c);
extern long   make_long    (const Cor_CArg& c);
extern float  make_float   (const Cor_CArg& c);
extern double make_double  (const Cor_CArg& c);
extern char  *make_string  (const Cor_CArg& c);
extern void  *make_chunk   (Cor_CArg *c);

extern char *functor_name    (const Cor_CArg& c);
extern Cor_ArgList *functor_args (Cor_CArg *c);

extern Cor_CArg *make_car (const Cor_CArg& c);
extern Cor_CArg *make_cdr (const Cor_CArg& c);

// **************************************************************************
//    Interface functions for finding a database relation:

extern Cor_Relation *find_relation (char *db_rel_name, int arity);

// **************************************************************************
//    Interface functions for calling the declarative module:

extern int call_coral(char *exp_pred_name, Cor_Tuple *query,
        Cor_Relation *&result);
    //
    // The query is a tuple for the relation, and any
    // facts in the relation that unify with the query tuple are 
    // answers to the query.
    // For example, (X,Y) is a query that returns all tuples in a binary
    // relation.
    // Similarly, (1,Y) returns all tuples that have 1 in the first 
    // column.
    // Adds answer tuples to the relation result
    // If the result parameter is NULL, a new relation is
    // allocated and result points to it.
    // Returns:  Number of answer tuples.  Result < 0 implies error.

extern int call_coral(char *exp_pred_name, Cor_Relation *queries,
              Cor_Relation *&result);
    //
    // Relation *queries must contain queries on the relation.
    // Adds answer tuples to the relation result.
        // If the result parameter is NULL, a new relation is
        // allocated and result points to it.
    // Returns:  Number of answer tuples.  Result < 0 implies error.

// **************************************************************************
//    Interface functions for stepping through and modifying a relation:

/** FOR_EACH_TUPLE(Cor_Tuple *tuple, Cor_Relation *rel) **/

#define FOR_EACH_TUPLE(tuple,rel) {\
    Cor_StackMark CONCAT_ID2(tuple,_stack_mark);\
    Cor_BindEnv *CONCAT_ID2(tuple,_env) = \
        new Cor_BindEnv((rel)->arity()); \
    Cor_ArgList *alist0 = make_vararglist((rel)->arity()); \
    Cor_TupleIterator CONCAT_ID2(ITER,tuple) (rel, *alist0, \
        CONCAT_ID2(tuple,_env)); \
    for(;;) {    \
        (tuple) = CONCAT_ID2(ITER,tuple).get_next_tuple(); \
        if (CONCAT_ID2(ITER,tuple).no_match()) break;    

#define END_EACH_TUPLE(tuple)  } \
    CONCAT_ID2(tuple,_stack_mark).pop_to();\
    delete CONCAT_ID2(tuple,_env); \
    delete alist0; }

// *****************************

/** FOR_EACH_MATCHING_TUPLE(Cor_Tuple *tuple, Cor_Relation *rel,
            Cor_Tuple *match_tuple) **/

#define FOR_EACH_MATCHING_TUPLE(tuple,rel,match_tuple) {\
    Cor_StackMark CONCAT_ID2(tuple,_stack_mark);\
    Cor_BindEnv *CONCAT_ID2(tuple,_env) = \
        new Cor_BindEnv((rel)->arity()); \
    Cor_TupleIterator CONCAT_ID2(ITER,tuple) \
        (rel, (match_tuple)->args(), CONCAT_ID2(tuple,_env)); \
    for(;;) {       \
        (tuple) = CONCAT_ID2(ITER,tuple).get_next_tuple(); \
        if (CONCAT_ID2(ITER,tuple).no_match()) break;       

// **************************************************************************

extern Cor_Tuple *make_tuple(int arity);
extern Cor_Tuple *make_tuple(Cor_ArgList *arglist);

extern Cor_Relation *make_rel(char *rel_name, int arity);


// **************************************************************************
// Interface functions for printing

/** Cor_CArg::printon(CoralStream *file);  and 
    Cor_ArgList::printon(CoralStream *file);

    Cor_Tuple::printon(CoralStream *file);
    Cor_Relation::printon(CoralStream *file);

**/

// **************************************************************************
//    Interface functions for creating indices for relations:

// **************************************************************************
//    Interface functions for calling the imperative module:

/** The user needs to write a routine with signature
    Cor_Relation *pred_solver (Cor_Tuple *query);    or
    Cor_Tuple *pred_solver (Cor_Tuple *query);
    for the predicate `pred', and add a declaration matching the form:
    int add_builtin (char *pred, int arity, UserRelProc pred_solver);  or
    int add_builtin (char *pred, int arity, UserTupleProc pred_solver);
    
    where `arity' is the arity of the relation named by `pred'.
**/

extern int add_builtin (char *pred, int arity, UserRelProc pred_solver);
extern int add_builtin (char *pred, int arity, UserTupleProc pred_solver);
#define BUILTIN_FAIL_RETURN {__CORAL_BUILTIN_FAIL=1;return 0;}
#define VOID_BUILTIN_FAIL_RETURN {__CORAL_BUILTIN_FAIL=1;return;}

// **************************************************************************

#endif _Interface_H_

