// *************************************************************************
//    Util.C
// *************************************************************************

#include <stdio.h>
#include <iostream.h>
#include <fstream.h>

#include <unistd.h>
#include <stdlib.h>

#include "Util.h"

#include "CoralError.h"
#include "Globals.h"
#include "HashVal.h"
#include "Interface.h"
#include "HashSimpleRelation.h"
#include "Debug.h"

  // From CArg.y
extern "C" int CArg_parse();

// *************************************************************************
int file_exists(char *name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // This doesn't seem to work with streams?

    FILE *fp;

    if ((fp = fopen(name,"r")) == NULL)
        return(0);

    fclose(fp);
    return(1);
}

// *************************************************************************
CoralStream *Cor_open_infile(char *name, int mode)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ifstream *fs = NULL;
    CoralStream *tmp_file = NULL;

    if (!file_exists(name))
        return(NULL);

    if (name)
        fs = new ifstream(name,mode);

    if (!*fs){
        *(Cor_exEnv->err_strm) << "CORAL: unable to open file " << name << endl;
        delete fs; 
        return(NULL);
    } else
        tmp_file = new CoralStream(fs->rdbuf());

    if (tmp_file) {
        tmp_file->ifs = fs;
        tmp_file->orig = OST_IFS;
    }
    
    return(tmp_file);
}

// **************************************************************************
CoralStream *Cor_open_file(char *name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return(Cor_open_outfile(name));
}

CoralStream *Cor_open_outfile(char *name, int mode)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ofstream *fs = NULL;
    CoralStream *tmp_file = NULL;


    if (name)
        fs = new ofstream(name,mode);

    if (!fs) 
       *(Cor_exEnv->err_strm) << "CORAL: unable to open file " << name << endl;
    else
        tmp_file = new CoralStream(fs->rdbuf());
    
    if (tmp_file) {
        tmp_file->ofs = fs;
        tmp_file->orig = OST_OFS;
    }
    
    return(tmp_file);
}

// *************************************************************************
//    Reads in a specified file and dumps it on exEnv->err_strm.  If there are
//    any problems opening the file, 1 is returned, else 0 is returned.
// *************************************************************************
int Cor_DisplayFile(char *file_name, int use_pager)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char file_buffer[1000];
    char buff[1000];
    FILE *fp;
      
    ASSERT((strlen(file_name)<1000));

    // Quick check that file exists
    if ((fp = fopen(file_name,"r")) < 0)
        return(1);
    else
        fclose(fp);

    if (coralConfig->server_mode || !use_pager) {
          // Read the file and dump it into the socket
        CoralStream *strm = Cor_open_infile(file_name);
        if (strm == NULL)
            return(1);
        while(strm->gets(buff, 1000))
            *(Cor_exEnv->out_strm) << buff;
        delete strm;
        return(0);
    } else {
        // Should probably use $PAGER instead or more.
        sprintf(file_buffer, "more %s", file_name);
    }

    return(system(file_buffer));
}

// *************************************************************************
//   Using the keyword given, find the base file name of the help file it's
//   in.  If the keyword doesn't match any in the coral_help relation, 
//   return NULL.  Also returns the line number the database has if
//   ln_num is not NULL.
// *************************************************************************
char *Cor_FindHelpFile(Cor_CArg *keywrd, int *ln_num)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char *basename = NULL;

    if (!Cor_HelpRel)
        return(NULL);

    if (! keywrd->is_symbol())
        return(NULL);

    Cor_ArgList  *args = make_vararglist(3);
    (*args)[0] = *keywrd;
    Cor_Tuple    *mtch_tuple = new Cor_Tuple(args);
    Cor_Tuple    *tuple = NULL;

    Cor_ScanDesc *scan = new Cor_ScanDesc(Cor_HelpRel,mtch_tuple);

    if (scan) {
          // Should be at most one matching fact        
        tuple = scan->next_tuple();
        if (!scan->no_match()) {
            if ( (*tuple)[1].is_symbol() && (*tuple)[2].is_int() ) {
                basename = (*tuple)[1].symbol(); 
                if (ln_num)
                    *ln_num = (*tuple)[2].NumIval();  
            }
        }
    }

    delete scan;
    delete mtch_tuple;
    delete args;

    return(basename);
}

// *************************************************************************
int Cor_consultFile(char *filename, char *stack_filename)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    CoralStream *tmp_strm;

    tmp_strm = Cor_open_infile(filename);

    if (tmp_strm == NULL)
        return(0);
    else {
        Cor_parseStack->handle_new_file(tmp_strm, stack_filename);
        // Breaks RODIN -spf
        //if (stack_filename) {
            //char temp_str[200];
            // Remove the <filename>.M file
            //sprintf(temp_str, "%s.M", stack_filename);
            //unlink(temp_str);
        //}
    }

    return(1);
}

// *************************************************************************
int Cor_consultFile(CoralStream *coral_in, char *stack_filename)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    Cor_parseStack->handle_new_file(coral_in, stack_filename);

    if (stack_filename) {
        char temp_str[200];
        // Remove the <filename>.M file
        sprintf(temp_str, "%s.M", stack_filename);
        unlink(temp_str);
    }

    return(1);
}

// *************************************************************************
int Cor_consultBuff(char *buffer, char *buffname)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!buffer || (! *buffer))
        return(0);
    else {
        if (buffname) {
            char temp_str[200];
            // Remove the <buffname>.M file
            sprintf(temp_str, "%s.M", buffname);
            unlink(temp_str);
        }

        Cor_parseStack->handle_new_buff(buffer,buffname);
            /* Can safely do this in server mode? */
        if (coralConfig->server_mode)
            Cor_parseStack->parse(Cor_exEnv->out_strm,1);
    }
    return(1);
}

// *************************************************************************
char *Cor_mk_tmp_file_name(char *name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    static filecount = 0;
    sprintf(name, "/tmp/__coral_file_%d", filecount++);
    return (name);
}

// *************************************************************************
char *Cor_make_done_prefix()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return "done_";
}

// *************************************************************************
//  Convert an adornment of the form "bbf" to a BitVector 110.  Return 0 if
//  ok, -1 otherwise.
// *************************************************************************
int Cor_ConvertAdornment(Cor_BitVector& bits,
            char *adorn_string, int adorn_length)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    for (int i = 0; i < adorn_length; i++) {
        char adorn_char = *adorn_string++;
        switch (adorn_char) {
        case 'b':
            bits.setlen(i+1);
            bits.set(i, 1);
            break;
        case 'f':
            bits.setlen(i+1);
            bits.set(i, 0);
            break;
        default:
            return(-1);
        }
    }
    return(0);
}

// *************************************************************************
Cor_Tuple *Cor_parseFact(char *fact_str, Cor_CArg& rel_name, int *arity)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
  // These items should be defined in CArg.y - the mini parser
extern Cor_CArg *Cor_RelName;
extern Cor_CArg *Cor_retCArg;
extern int       Cor_RelArity;

int num_args = 0, num_needed = 0;
Cor_ArgList *al  = NULL;
Cor_Tuple *tuple = NULL;

    if (Cor_RelName)
        num_needed = Cor_RelArity;

    if (fact_str)
        Cor_consultBuff(fact_str, "pf_buff");

    if (Cor_parseStack->is_empty())
        return(NULL);

    do {

        CArg_parse();

        if (Cor_retCArg) {

            // Got a new Arg

            if (Cor_RelName && !Cor_RelName->is_empty()) {

                  // This is a table file

                 if (!al)
                     al = new Cor_ArgList(Cor_RelArity);

                (*al)[num_args++] = *Cor_retCArg;
                Cor_retCArg->destroy();
                Cor_retCArg = NULL;

            } else {

                  // This appears to be a regular fact file, so we have an
                  // entire fact in Cor_retCArg
       
                if (Cor_retCArg->is_functor()){
   
                    Cor_CArg *farg = Cor_retCArg;

                   *arity = farg->arity();
                    rel_name = Cor_CArg(farg->functor());
        
                      // need to make copy of ArgList
                    al = (farg->args()).dup();
      
                    delete Cor_retCArg;
                    Cor_retCArg = NULL;

                    break;
                } else {
                    // Not a functor
                    // Could be a special symbol "quit", etc.
                    rel_name = *Cor_retCArg;
                    Cor_retCArg->destroy();
                    Cor_retCArg = NULL;
                    break;
                }
            }

        } else if (Cor_RelName) {
          // Relation Name and arity have changed (for table file)

            rel_name = *(Cor_RelName);
           *arity = num_needed = Cor_RelArity;

        } else {
            // EOF or error
            if (num_args) {
                *(Cor_exEnv->err_strm)
                        << "Parse Error: Only got a partial fact.\n";
            }
            return(NULL);
        }

    } while (num_args < num_needed);

   
    if (al) 
        tuple = new Cor_Tuple(al);

      // Only one fact per string?
    if (fact_str)
        Cor_parseStack->get_next_file();

      // Only do this for table files
    if (Cor_RelName && (! Cor_RelName->equals(rel_name))) {
        rel_name = *(Cor_RelName);
       *arity = Cor_RelArity;
    }

    return(tuple);
}

// *************************************************************************
// This routine takes the name of the table/fact file in, and processes it,
// returning a relation.  It does _not_ enter the relation into the 
// internal tables, etc.
// *************************************************************************
Cor_Relation *Cor_ReadRelation(char *infile_nme, int arity)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!Cor_consultFile(infile_nme, infile_nme)) {
        CORAL_error(COR_FILE_NOT_FOUND, infile_nme, "ReadRelation");
        return(NULL);
    }

    extern Cor_CArg *Cor_retCArg;
    extern Cor_CArg *Cor_RelName;

    if (Cor_retCArg) {
        Cor_retCArg->destroy();
        Cor_retCArg = NULL;
    }

    if (Cor_RelName) {
        Cor_RelName->destroy();
        Cor_RelName = NULL;
    }

    // Now read in the lines using parseFact() and insert into the relation
    Cor_HashSimpleRelation *rel = new Cor_HashSimpleRelation(arity);
    Cor_Tuple *tuple = NULL;

    Cor_CArg relnme;
    int rel_arity = 0;

    tuple = Cor_parseFact(NULL, relnme, &rel_arity);

    while (tuple) {

        // If its a new relation find or create it
        if (rel_arity == arity) {

              // Insert fact into Relation rel
            if (! rel->insert_tuple(tuple))
                delete tuple;
        } else {
            // Ignore incorrect fact
            delete tuple;
        }

        tuple = Cor_parseFact(NULL, relnme, &rel_arity);
    }

    return(rel);
}

// *************************************************************************
//   CheckArgTypes:
//
//   This function type checks the argument list given against the 
//   template provided, as well as checking that the minimum and 
//   maximum number of arguments are satisfied.  A max of -1 indicates
//   that an arbitrary number of arguments are allowed, they are all
//   required to be the same type as the last type mentioned in the type list.
//   If the terms array exists and copy_term is non-zero, the array will
//   have the terms placed into it, otherwise, if the array is present,
//   it is considered to already hold the terms.
//   (Useful for checking against multiple possiblities.)
//
//   A non-null env passed in indicates the terms should
//   be deep_dereferenced using that BindEnv.
//
//   Acceptable types:
//
//        a    - anything goes (don't check this arg)
//        i    - integer
//        d    - double
//        n    - numeric (int or double)
//        f    - functor
//        l    - list
//        s    - symbol
//        v    - a variable (unbound, if arguments were dereferenced)
//
//  A few examples:   min    max    list        meaning
//                     2      5     "sln"       1st arg: symbol
//                                              2nd arg: list
//                                              3rd->5th: numeric (if present)
//
//                     0     -1     "s"         Arbitrary number of args,
//                                              all present must be symbols.
//                     2      2     "fv"        Takes exactly 2 arguments,
//                                              a functor and a variable.
//
// *************************************************************************
Cor_ArgType_Error Cor_CheckArgTypes(Cor_ArgList& args,
        int min_args, int max_args, char *tlist,
        Cor_Term *terms,
        int copy_terms, Cor_BindEnv *env )
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int  j, cnt = args.count();

    if (cnt < min_args)
        return(Coral_Not_Enough_Args);

    if (max_args >= 0)
        if (cnt > max_args)
            return(Coral_Too_Many_Args);

      // Copy the args to the array supplied
    if (terms && copy_terms) 
        for (j=0; j < cnt; j++) {
            terms[j].set(args[j], env);
            if (env)
                Cor_deep_dereference_term(terms[j]);
        }

      // No type checking is desired
    if (!tlist || !*tlist)
        return(Coral_Args_Okay);

      // Check for argument types
    char *type = tlist;

    if (terms) {
        // Use the dereferenced terms
        for (j=0; j < cnt; j++) {
    
            switch (*type) {
    
            case 'n': 
                if (!terms[j].expr.is_num())
                    return(Coral_Wrong_Type_Args);
                break;
    
            case 'i':
                if (!terms[j].expr.is_int())
                    return(Coral_Wrong_Type_Args);
                break;
    
            case 'd':
                if (!terms[j].expr.is_double())
                    return(Coral_Wrong_Type_Args);
                break;
    
            case 'f':
                if (!terms[j].expr.is_functor())
                    return(Coral_Wrong_Type_Args);
                break;
    
            case 'l':
                if (!terms[j].expr.is_list())
                    return(Coral_Wrong_Type_Args);
                break;
    
            case 's':
                if (!terms[j].expr.is_symbol())
                    return(Coral_Wrong_Type_Args);
                break;
    
            case 'v':
                if (!terms[j].expr.is_var())
                    return(Coral_Wrong_Type_Args);
                break;
    
            case 'u':
                if (!terms[j].expr.is_var())
                    return(Coral_Wrong_Type_Args);
                if (env && !env->not_bound(terms[j].expr.var_num()))
                    return(Coral_Wrong_Type_Args);
                break;
    
            default:
                TELL2("Bad type-string argument passed in to ");
                ASSERT(0);
            case 'a':    // Any - no check performed
                break;
            }
    
            if (! *++type)
                type--;
        }

    } else {
        // Use the arglist
        for (j=0; j < cnt; j++) {
    
            switch (*type) {
    
            case 'n': 
                if (!args[j].is_num())
                    return(Coral_Wrong_Type_Args);
                break;
    
            case 'i':
                if (!args[j].is_int())
                    return(Coral_Wrong_Type_Args);
                break;
    
            case 'd':
                if (!args[j].is_double())
                    return(Coral_Wrong_Type_Args);
                break;
    
            case 'f':
                if (!args[j].is_functor())
                    return(Coral_Wrong_Type_Args);
                break;
    
            case 'l':
                if (!args[j].is_list())
                    return(Coral_Wrong_Type_Args);
                break;
    
            case 's':
                if (!args[j].is_symbol())
                    return(Coral_Wrong_Type_Args);
                break;
    
            case 'v':
                if (!args[j].is_var())
                    return(Coral_Wrong_Type_Args);
                break;
    
            default:
                TELL2("Bad type-string argument passed in to ");
                ASSERT(0);
            case 'a':    // Any - no check performed
                break;
            }
    
            if (! *++type)
                type--;
        }
    }

    return(Coral_Args_Okay);
}

// *************************************************************************
