// *************************************************************************
//    DerivedMethod.C
// *************************************************************************

#include <stdio.h>

#include "DerivedMethod.h"

#include "Config.h"
#include "DerivedRelation.h"
#include "Database.h"
#include "Unparse.h"
#include "Predicate.h"
#include "PipelinedDerivedRelation.h"
#include "QFModuleData.h"
#include "Debug.h"

// *************************************************************************
Cor_DerivedMethod::Cor_DerivedMethod(Cor_CArg& name, Cor_BitVector& adorn,
    InterpDerivedSolver solvr, Cor_QFModuleData *module)
    : adornment(adorn)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_DerivedRelation *rel;

    valid_method = 1;
    md = module;
    if (Cor_CurDB->RelationTable == NULL) {
        Cor_CurDB->RelationTable = new Cor_SymTable();
    }

    char buf[1000];

    strcpy(buf, name.symbol());
    sprint_adornment(buf+strlen(buf), &adorn);
    Cor_CArg check_name = Cor_CArg(buf);
    
    if (Cor_is_magic(check_name) || Cor_is_supp(check_name)
            || Cor_is_mp_done(check_name))
    {
        *(Cor_exEnv->err_strm) << "Warning::relation name "
            << buf << " conflicts with a system reserved name.\n";
        *(Cor_exEnv->err_strm) << "New definition not retained.\n";
        return;
    }
    else {
        sprintf(buf, "%s", name.symbol());
    }
    
    Cor_CArg new_name = Cor_CArg(buf);

    Cor_Relation *reln = (Cor_Relation *)
            Cor_CurDB->RelationTable->lookup(new_name);

    if (! reln) {
        if (module->kindof() == COR_PIPELINED) {
            reln = new
                Cor_PipelinedDerivedRelation((Cor_QFPipelinedModuleData *) md,
                new_name, adorn.len());
            md = NULL;
            // The md field is not used by pipelined relations. Instead
            // the ModuleData is stored along with the relation

        } else {
            reln = new Cor_DerivedRelation(new_name, adorn.len());
        }

        reln->single_success = module->module_info->SingleAnswer;
        Cor_CurDB->RelationTable->insert(new_name, (Pointer) reln);

    } else {

          // search methods already defined to see if there is a conflict
        if (reln->r_kind != COR_R_DERIVED) {
            *(Cor_exEnv->err_strm) << "Base relation "
                    << new_name.symbol() << " already exists.\n";
            *(Cor_exEnv->err_strm) << 
                    "Newer definition is NOT retained.\n";
            return;
        }

        rel = (Cor_DerivedRelation *) reln;
        Cor_DerivedMethod *meth = rel->methods;

        //  Three cases: 
        //  1)  The new definition is a pipelined relation.  In this case, we
        //      have to first delete the original relation, and insert the
        //      new one
        //  2)  The new definition is a Derived relation, and the old
        //      definition is a pipelined relation.  In this case too, we have
        //      to first delete the original relation, and insert the new one.
        //  3)  Both definitions are Derived relations.  Only the ModuleData
        //      need be replaced

        if (!meth->md || (module->kindof() == COR_PIPELINED)) {
            *(Cor_exEnv->err_strm) << "Duplicate definition of relation "
                    << new_name.symbol() << ".\n";
            *(Cor_exEnv->err_strm) << "Retaining newer definition.\n";

            rel = (Cor_DerivedRelation *)
                    Cor_CurDB->RelationTable->delete_entry(rel->name);
              // delete the old relation, and all its methods
            delete rel;

            if (module->kindof() == COR_PIPELINED) {
                reln = (Cor_Relation *) new Cor_PipelinedDerivedRelation(
                    (Cor_QFPipelinedModuleData*)md, new_name, adorn.len());
                md = NULL;
                // This field is not used by pipelined relations. Instead
                // the ModuleData is stored along with the relation 

            } else {
                reln = (Cor_Relation *)
                        new Cor_DerivedRelation(new_name, adorn.len());
            }

            reln->single_success =
                module->module_info->SingleAnswer;

            Cor_CurDB->RelationTable->insert(new_name, reln);
            rel = (Cor_DerivedRelation *) reln;
            next = rel->methods;
            rel->methods = this;
            interp_solver = solvr;
            return;
        }

        while (meth) {
            if (meth->valid_method && (meth->adornment==adorn)) {
                *(Cor_exEnv->err_strm) << 
                    "Duplicate definition with identical adornment.\n";
                *(Cor_exEnv->err_strm) << "Retaining newer definition of "
                        << new_name.symbol();
                Cor_print_adornment(adorn, Cor_exEnv->err_strm);
                *(Cor_exEnv->err_strm) << ".\n";

                meth->interp_solver = solvr;
                if (meth->md)
                    delete meth->md;
                meth->md = md;
                return;
            }
            meth = meth->next;
        }
    }

    rel = (Cor_DerivedRelation *) reln;
    next = rel->methods;
    rel->methods = this;
    interp_solver = solvr;
}

// *************************************************************************
Cor_DerivedMethod::~Cor_DerivedMethod()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (md) delete md;
}

// *************************************************************************
