// *************************************************************************
//    RelationMisc.C
// *************************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "RelationMisc.h"

#include "Globals.h"
#include "CoralError.h"

#include "Database.h"
#include "DatabaseStruct.h"
#include "QFModuleData.h"
#include "Literal.h"
#include "Relation.h"
#include "StorageRelation.h"
#include "AggSelInfo.h"
#include "Execute.h"
#include "HashSimpleRelation.h"
#include "Debug.h"

// *************************************************************************

#define MAX_STORED_TUPLEITS 1024

// *************************************************************************

int Cor_RelCount = 0;

// **************************************************************************

extern int Cor_subsumes_arg_lists(Cor_ArgList& left, Cor_ArgList& right,
    Cor_BindEnv *left_bindings, Cor_BindEnv *right_bindings);

// *************************************************************************
// Returns the basename, and gives pointers for magic and
// adornments (or NULLs).  The relname buffer passed in can be modified.
char *Cor_split_relname(char *relname, char **magic, char **adorn)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char *base;

    if (strncmp(relname,"m_",2) == 0) {
        if (magic)
           *magic = relname;
        base  = relname + 2;
    } else {
        if (magic)
           *magic = NULL;
        base  = relname;
    }

    char *adrn = strrchr(base,'_');

    if (adrn) {
        int is_adorn = 1;

        for (char *ch=adrn+1; *ch; ch++)
            if ((*ch != 'b') && (*ch != 'f')) {
                is_adorn = 0;
                break;
            }

        if (is_adorn) {
            *adrn = '\0';
            if (adorn)
                *adorn = adrn+1;

        } else
            if (adorn)
                *adorn = NULL;
    }

    return(base);
}

// *************************************************************************
void Cor_group_dump(Cor_BindEnv *new_env, Cor_ArgList *new_arglist,
        Cor_Relation *rel, char *old_arglist)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (Cor_T_Stack->count <= 0 || Cor_T_Stack->top() == NULL)
        return;
    
    Cor_Table *F = Cor_T_Stack->top();
    
    Cor_RuleInfo * rInfo = F->moduledata->curr_rInfo;
   
    ostream *strm = Cor_exEnv->explain_strm; 
    if (!strm)
        return;

    for (int j=0; j <= rInfo->num_literals; j++) {
        if (j==rInfo->num_literals) {
            // First Argument
            *strm << "def_" << F->moduledata->export_pred->name.symbol()
                  << "(";
            rInfo->rule->head->predicate->name.print(new_env,strm);
            *strm << "(";
            ((Cor_ArgList *)new_arglist)->print_dump(new_env,strm);
            
            *strm << ")";
        }
        else {
            *strm << "use_" << F->moduledata->export_pred->name.symbol()
                << "(";
            rInfo->rule->preds[j]->predicate->name.print(new_env,strm);
            *strm << "(";
            rInfo->arg_list[j].print_dump(new_env,strm);
            *strm << ")";
        }
            
        *strm << ", [";
        
        rInfo->rule->head->predicate->name.print(new_env,strm);
        *strm << "(";
        ((Cor_ArgList *)new_arglist)->print_dump(new_env,strm);
        *strm << ")";
        
        for (int i=0; i < rInfo->num_literals; i++) {
            *strm << ",";
            rInfo->rule->preds[i]->predicate->name.print(new_env,strm);
            *strm << "(";
            rInfo->arg_list[i].print_dump(new_env,strm);
            *strm << ")";
        }
        if (old_arglist) {
            *strm << ", " << rel->name.symbol() << "_SUBSUMED__";
            *strm << "(" << old_arglist << ")";
        }
            
        *strm << "]).\n";
    }


    // For two new relations: def_new and use_new
    for (int j1=0; j1 <= rInfo->num_literals; j1++) {
        if (j1==rInfo->num_literals) {
            // First Argument
            *strm << "def_new_" << F->moduledata->export_pred->name.symbol()
                << "(";
            rInfo->rule->head->predicate->name.print(new_env,strm);
            *strm << "(";
            ((Cor_ArgList *)new_arglist)->print_dump(new_env,strm);
            *strm << ")";
        } else {
            *strm << "use_new_" << F->moduledata->export_pred->name.symbol()
                << "(";
            rInfo->rule->preds[j1]->predicate->name.print(new_env,strm);
            *strm << "(";
            rInfo->arg_list[j1].print_dump(new_env,strm);
            *strm << ")";
        }

        *strm << ", [";

        rInfo->rule->head->predicate->name.print(new_env,strm);
        *strm << "(";
        ((Cor_ArgList *)new_arglist)->print_dump(new_env,strm);
        *strm << ")";

        for (int i1=0; i1 < rInfo->num_literals; i1++) {
            *strm << ",";
            rInfo->rule->preds[i1]->predicate->name.print(new_env,strm);
            *strm << "(";
            rInfo->arg_list[i1].print_dump(new_env,strm);
            *strm << ")";
        }
        if (old_arglist) {
            *strm << ", " << rel->name.symbol() << "_SUBSUMED__";
            *strm << "(" << old_arglist << ")";
        }

        *strm << "]).\n";
    }

    *strm << flush;
}

// *************************************************************************
RelationKind Cor_find_r_kind(Cor_CArg& name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (Cor_is_magic(name))
        return COR_R_MAGIC;
    else if (Cor_is_supp(name))
        return COR_R_SUPPLEMENTARY;
    else
        return COR_R_ANSWER;
}

// *************************************************************************
/*
    *::release(TupleIterator &iterator)
  
    Called when the tupleiterator is no longer needed, and the state
    stored in the tuple iterator can be released.  The get_next / get_next_tuple
    functions store data in the tuple iterator, so that the data is saved
    across calls to get_next / get_next_tuple.  When the external routine is
    done with getting new answers, it calls tuple_iterator.release(), 
    which in turn calls reln::release(tuple_iterator).  

    WARNING:: release may get called more than once with the 
    same tuple_iterator.  Care should be taken to avoid deleting space
    more than once.  This is typically done by resetting pointed fields 
    to NULL after deletion.
*/

// *************************************************************************
//  checks the relation to see if the anno has already been specified
//  temporary kludge !! - PRAVEEN
// *************************************************************************
int Cor_duplicate_agg_sel_anno(Cor_StorageRelation *rel,
            Cor_AggSelInfo *newinfo)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_AggSelInfo *aggsel;
         
    for (aggsel=rel->agg_sel_info; aggsel; aggsel=aggsel->next) {
        if ((newinfo->agg_kind == aggsel->agg_kind)
                && (newinfo->agg_val == aggsel->agg_val)
                && (newinfo->var_count == aggsel->var_count)) {
            return(1);
        }
    }

    return(0);
}

// *************************************************************************
Cor_Relation *Cor_find_local_or_external_relation(Cor_CArg& name, int arity) 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Relation *rel = Cor_find_local_relation(name, arity);

    if (!rel)
        rel = Cor_make_relation(name, arity);

    return(rel);
}

// *************************************************************************
Cor_Relation *Cor_find_external_relation(Cor_CArg& name, int /* arity */ )
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Relation *rel = (Cor_Relation *)
            Cor_CurDB->RelationTable->lookup(name);

    if (! rel) {
        rel = (Cor_Relation *)
                Cor_BuiltinDB->RelationTable->lookup(name);
    }

    return(rel);
}

// *************************************************************************
Cor_Relation *Cor_Find_External_Relation(Cor_CArg& name, int arity)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // This version creates a relation if none is found.
    // It gives a warning if it did not find an existing relation

    if (name.is_empty())
        return(NULL);

    Cor_Relation *rel = Cor_find_external_relation(name, arity);
    if (rel)
        return rel;
    
    if (!Cor_exEnv->C_quiet_mode_default) {
        CORAL_error(COR_REL_NOT_FOUND,name.symbol(),
                "Find_External_Relation");
        *(Cor_exEnv->trace_strm) << "Arity of relation = " << arity
                << "\nAllocating new relation.\n";
    }
    
    return Cor_make_relation(name, arity);
}

// *************************************************************************
Cor_Relation *find_relation(char *name, int arity) 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg n = Cor_CArg(name);
    return find_relation(n, arity);
}

// *************************************************************************
Cor_Relation *find_relation(Cor_CArg& name, int arity)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Relation *rel;

    rel = Cor_find_local_relation(name, arity);
    if (!rel)
        rel = Cor_find_external_relation(name, arity);

    if (!rel) {
        CORAL_error(COR_REL_NOT_FOUND,name.symbol(),"find_relation");
        *(Cor_exEnv->trace_strm)  << "Arity of relation = " << arity << endl;
    }
    return rel;
}

// *************************************************************************
Cor_Relation *Cor_make_relation(Cor_CArg& name, int arity, int delta_indexed)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char buf[400];

    if (name.is_empty())
        return(NULL);

    if (Cor_is_magic(name) || Cor_is_supp(name) || Cor_is_mp_done(name)) {
        *(Cor_exEnv->err_strm) << "Warning::relation name "
                << name.symbol()
                << " conflicts with a system reserved name.\n";
        sprintf(buf, "new%s%d", name.symbol(), arity);
        *(Cor_exEnv->err_strm) << "New name is " << buf << ".\n";
    } else
        sprintf(buf, "%s", name.symbol());

    Cor_CArg new_name = Cor_CArg(buf);

    Cor_Relation *rel = (Cor_Relation *)
            Cor_CurDB->RelationTable->lookup(new_name);

    if (! rel) {

        rel = (Cor_Relation *)
                Cor_BuiltinDB->RelationTable->lookup(name);

        if (! rel) {
            rel = Cor_AllocateRelation(new_name, arity, delta_indexed);
            Cor_CurDB->RelationTable->insert(new_name, rel);
        }
    }

    return(rel);
}

// *************************************************************************
Cor_Relation *Cor_make_relation(char *name, int arity, int delta_indexed)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg n = Cor_CArg(name);
    return Cor_make_relation(n, arity, delta_indexed);
}

// *************************************************************************
Cor_StorageRelation *Cor_make_local_relation(Cor_CArg& name, int arity,
    int delta_indexed, int set_subsum)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // this stuff ensures that the relation constructor sees the desired
    // value of exEnv->C_check_subsum_default    -- P

    if (set_subsum == -1)
        return Cor_AllocateRelation(name, arity, delta_indexed);


    int temp = Cor_exEnv->C_check_subsum_default;
    Cor_exEnv->C_check_subsum_default = set_subsum;
    Cor_StorageRelation *rel = Cor_AllocateRelation(name,arity,delta_indexed);
    Cor_exEnv->C_check_subsum_default = temp;
    return(rel);
}

// *************************************************************************
void Cor_reclaim_relation(Cor_StorageRelation *rel)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    delete rel;
}

// *************************************************************************
void Cor_reclaim_index(Cor_GenericIndex *index, RelationKind /* r_kind*/)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    delete index;
}

// *************************************************************************
// **************************************************************************
void delete_id_tuple(Cor_Relation *rel, Cor_Tuple *tuple)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#   ifdef DO_TRACE
    if (Cor_exEnv->GlobalRelationOptions & REL_DISPLAY_INSERTIONS) {
        *(Cor_exEnv->trace_strm) << "[Deleting "
                << rel->name.symbol() << "(";
        tuple->args().print(NULL,Cor_exEnv->trace_strm);
        *(Cor_exEnv->trace_strm) << ")]\n";
    }
#   endif

    int del_tup;

    if (tuple->is_in_tlist()) {

        // This must be a StorageRelation (of some kind)
        del_tup = ((Cor_StorageRelation *) rel)->tlist.delete_tuple(tuple);
    } else {
        del_tup = rel->get_index_set()->delete_tuple(tuple);
    }

      // Delete is only meaningful for StorageRelations
    Cor_StorageRelation *srel = (Cor_StorageRelation*) rel;
    if (!del_tup && srel->delta_indexed && srel->delta_relations) {

        // We just mark the tuple as deleted. The actual deletion from
        // the delta is done in a lazy fashion, when the tuple is 
        // moved to the main relation.

        tuple->do_delete();
        tuple->set_del_ref(
            srel->delta_relations->
                reln_chain->reln->get_index_set()->num_indices());

    } else {

        if (!rel->is_borrow()) {
            if (!tuple->has_no_args())
                delete tuple->_args;

            delete tuple;
        }
    }

    (rel->count)--;
}

// **************************************************************************
//  This is a value based delete, where every tuple subsumed by tuple is
//  deleted from the relation. 
void delete_tuple(Cor_Relation *rel, Cor_Tuple *tuple)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int env_size;
    Cor_BindEnv *env;
    Cor_StackMark stack_mark;
    int check_subsum = rel->check_subsum;

    env_size = tuple->args().max_vars();

    if (env_size)
        env = new Cor_BindEnv(env_size);
    else
        env = NULL;

    Cor_TupleIterator it(rel, tuple->args(), env);

    for (;;) {

        Cor_Tuple *new_tuple = it.get_next_tuple();

        if (it.no_match())
            break;

        delete_id_tuple(rel, new_tuple);

        if ((check_subsum == 0) && (env_size == 0)) {
            break;
        }

        stack_mark.pop_to();
    }

    stack_mark.pop_to();

    delete env;
}

// **************************************************************************
//  update_tuple does not update indices yet.  Fix this bug sometime.
void update_tuple (Cor_Tuple *tuple, int arg_pos, Cor_CArg& new_val)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if ((arg_pos < 0) || (arg_pos >= tuple->args().count())) {

        *(Cor_exEnv->err_strm)
                 << "Error: update_tuple, argument position " << arg_pos
                 << " does not exist.\n";
        return /* 0 */ ;

      // the argument position exists
    } else {
        (tuple->args())[arg_pos] = new_val;
        return /* 1 */ ;
    }
}

// **************************************************************************
void update_tuple (Cor_Relation *rel, Cor_Tuple *old_tuple,
        Cor_Tuple *new_tuple)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    delete_tuple(rel, old_tuple);
    rel->insert_new(new_tuple);
}

// **************************************************************************
