// *************************************************************************
//    Rule.C
// *************************************************************************

#include "Rule.h"
#include "Literal.h"
#include "Globals.h"
#include "Debug.h"

#include "Unparse.h"
extern int Cor_print_new_line_after_rule;


// *************************************************************************
Cor_Clause::Cor_Clause()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    link = NULL;
    var_names = NULL;
    head = NULL;
    num_preds = 0;
    head_deletes_count = 0;
    head_deletes = NULL;
    preds = NULL;
    sips = NULL;
    line_number = -1;
    type = COR_NORMALRULE;
    seen = 0;
    recursive = 0;

    clause_names = new Cor_ClauseNames();
}

// *************************************************************************
Cor_Clause::~Cor_Clause()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int  j;

      // Magic does a lot of sharing with its Literals
    if (! Cor_exEnv->i_am_magic()) {

        delete head;

        for (j=0; j < num_preds; j++) {
            delete preds[j];
        }

        delete [] preds;

        for (j=0; j < head_deletes_count; j++)
            delete head_deletes[j];

        delete [] head_deletes;
    }

    delete var_names;
    delete clause_names;
}

// *************************************************************************
//  Makes a copy of the header fields of the rule, but not the head or body
//  predicates.  Should be called copy_proto_rule.

Cor_Clause *Cor_copy_rule(Cor_Clause *prototype)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Clause *rule;

    rule = new Cor_Clause();

    if (prototype) {

        delete rule->clause_names;
        rule->clause_names = prototype->clause_names->dup();

        if (prototype->var_names && (prototype->var_names->count() != 0)) {
            rule->var_names = prototype->var_names->dup();
            rule->type = prototype->type;
        }

        rule->line_number = prototype->line_number;
    }

    return(rule);
}

// *************************************************************************
void Cor_alloc_preds(Cor_Clause *rule, int count)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    rule->num_preds = count;
    rule->preds = new (Cor_Literal*)[count];
}

// *************************************************************************
ostream& operator<< (ostream& out, const Cor_Clause& rule)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j;
    int cn, nclauses;
    Cor_ClauseNames *cname = rule.clause_names;

    nclauses = cname->count();

    for (cn=0; cn < nclauses; cn++ ) {

            // Need to add quotes around labels
        out << "\"";
        out << cname->ith_symbol(cn);
        out << "\"";

        if (cname->ith_neg(cn))
            out << "-";
        out << ": ";
    }

    Cor_print_pred(rule.head, &out);

    for (int i=0; i < rule.head_deletes_count; i++) {
        out << ", ";
        Cor_print_pred(rule.head_deletes[i], &out, /* head */ 1);
    }

    if ((rule.num_preds > 0) || (rule.sips != NULL)) {
        out << " :- \n\t\t";
        for (j = 0; j < rule.num_preds; ++j) {
            Cor_print_pred(rule.preds[j], &out);
            if (j < rule.num_preds - 1)
                out << ", \n\t\t";     // print a pred on a line.
        }

        if (j==0)
            out << "true";             // print "true" if no rhs preds

        if (rule.sips) {
            out << " \n \t \t alias ";
            Cor_print_sips((Cor_Clause*) &rule, rule.sips, &out);
        }
    }
    out << ".";

    // Default is to print the new line. This is used by the magic program.
    // While dumping trace code, we dont want new lines printed. So then
    // the variable is cleared.
    if (Cor_print_new_line_after_rule)
        out << endl;

    return(out);
}

// *************************************************************************
