#ifdef DO_WODI

#include <fstream.h>

void print_cur_time(void);

#ifdef WODI_TIMEOUT

#include <sys/time.h>
#include "pvm3.h"

int wodi_trecvresponse(int, struct timeval *);

int wodi_trecvresponse(int tag, struct timeval * timeout)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    //    if (pvm_trecv(wodi_id, tag, timeout) < 0) return -1;    
    if (pvm_trecv(-1, tag, timeout) <= 0) return -1;    
    return pvm_getmwid(pvm_getrbuf());
}

#endif

//**************************************************************************
//    WodiStream.C
//
//    The Cor_Wodibuf Class - A buffer class for providing stream access
//    to Wodi communication. 
//
//**************************************************************************


#include "WodiStream.h"
#include <assert.h>
#include <stdio.h>

//**************************************************************************
//    Constructor for a Cor_wodibuf -
//
//**************************************************************************

Cor_wodibuf::Cor_wodibuf(BUF_TYPE _type)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#define w_out cout
#define w_err cerr        

    init();

    if ((type = _type) == WODI_MASTER ) {
        wodi_id = 0;
    }

}

//**************************************************************************
//    Destructor for a Cor_wodibuf -
//    The put area is synchronized. 
//    
//**************************************************************************

Cor_wodibuf::~Cor_wodibuf()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    sync();

    delete[] backup;
}

//**************************************************************************
//    Initialize the Cor_wodibuf -
//    The sizes are set and the buffer is allocated.
//
//**************************************************************************

void Cor_wodibuf::init()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    get_buffer_size = put_buffer_size = 4096;
    allocate();
}

//**************************************************************************
//    Check if the Cor_wodibuf is valid -
//    Return 1 if valid. Return 0 if invalid.
//
//**************************************************************************

int Cor_wodibuf::Valid()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

}

//**************************************************************************
//    Allocate the put and get areas -
//    Return 0 if success. Return EOF otherwise.
//
//**************************************************************************

int Cor_wodibuf::doallocate()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int buffer_size = get_buffer_size + put_buffer_size;
    char* cp = new char[buffer_size];
    
    backup = new char[buffer_size];
    total_len = 0;
    
    if( cp == NULL ) {
        return(EOF);
    }
    setb(cp, cp + buffer_size, 1);
    setg(cp, cp, cp - 1);
    setp(cp + get_buffer_size, cp + buffer_size);
    return 0;
}

//**************************************************************************
//    Synchronize the put area -
//    Return EOF is faliure occurs. 
//    
//**************************************************************************

int Cor_wodibuf::sync()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return overflow(EOF);
}

//**************************************************************************
//    Consume Characters in the put area -
//    The characters in the put area is sent to receiver (with rec_tid)
//    Return EOF is faliure occurs.
//
//**************************************************************************

int Cor_wodibuf::overflow(int c)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char * cp = pbase();
    streamsize len = pptr() - cp;

    if (len > 0) {
        if (type == WODI_MASTER) { // send the work to slaves
            pvm_initsend(PvmDataDefault);
            // send the length of the msg first
            if (pvm_pkint(&len, 1, 1) < 0) {
                pvm_perror("pvm_pkint");
                return EOF;
            }

            // send the real msg
            if (pvm_pkbyte((char*)cp, len, 1) < 0) {
                pvm_perror("pvm_pkbyte");
                return EOF;
            }
        
            time_t cal_time = time(NULL);
            char * s = new char[20];
            strftime(s, 20, "%m/%d %H:%M:%S",localtime(&cal_time));
            (w_out) << s;
            (w_out).flush();

            (w_out) << " about to WODI_SENDWORK with wodi_id = " << wodi_id << endl;
            (w_out).flush();
        
            wodi_sendwork(wodi_id);

            // modification for affinity
            //wodi_sendwork_with_class(wodi_id, (wodi_id % 5));
            
            strcpy(backup, cp);

            cal_time = time(NULL);
            strftime(s, 20, "%m/%d %H:%M:%S",localtime(&cal_time));
            (w_out) << s;
            (w_out).flush();
            delete s;

            (w_out) << " WODI_SENDWORK done with wodi_id = "<< wodi_id << endl;
            (w_out) << "QUERY = " << cp << endl; 
            (w_out).flush();
        }
        else { // send the result to master
            //            wodi_sendresponse(-1, CORAL_RESP_TAG);
            strncpy(backup + total_len, cp, len);
            total_len += len;
        }
    }
    
    setp(pbase(), epptr());
    if( c != EOF ) {
        return sputc(c);
    }
    return 0;
}

int Cor_wodibuf::wodi_flush()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    assert(type == WODI_SLAVE);
    pvm_initsend(PvmDataDefault);
    // send the length of the msg first
    if (pvm_pkint(&total_len, 1, 1) < 0) {
        pvm_perror("pvm_pkint");
        return EOF;
    }
    // send the real msg
    if (pvm_pkbyte((char*)backup, total_len, 1) < 0) {
        pvm_perror("pvm_pkbyte");
        return EOF;
    }
    print_cur_time(); cout << " about to WODI_SENDRESPONSE\n";cout.flush();
    wodi_sendresponse(-1, CORAL_RESP_TAG);
    print_cur_time();cout << " WODI_SENDRESPONSE done.\n";
    cout << "RESULT = " << backup << endl;
    cout.flush();
    total_len = 0;
    return 0;
}

//**************************************************************************
//    Supply characters for fetching -
//    Message from receiver (with rec_tid) is received and put to put area. 
//    Return EOF is faliure occurs.
//**************************************************************************

int Cor_wodibuf::underflow()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char* cp = gptr();
    if( egptr() > cp ) {        // chars still in buffer
        // gbump(1);  // Needed?
        return(*cp);
    }
    sync();                     // make sure our output is flushed

    // before we wait for input
    cp = eback();

    if (type == WODI_MASTER) { // receive the result from slave
#ifndef WODI_TIMEOUT

        time_t cal_time = time(NULL);
        char * s = new char[20];
        strftime(s, 20, "%m/%d %H:%M:%S",localtime(&cal_time));
        (w_out) << s;
        (w_out).flush();
        delete s;
        (w_out) << " about to WODI_RECVRESPONSE with wodi_id = "
                << wodi_id << endl; (w_out).flush();
        //  assert (wodi_id == wodi_recvresponse(CORAL_RESP_TAG));
        wodi_id = wodi_recvresponse(CORAL_RESP_TAG); 
#else
        struct timeval timeout;
        timeout.tv_sec = 60*40;
        timeout.tv_usec = 0;
        print_cur_time();
        cout << " about to WODI_TRECVRESPONSE with wodi_id = "
             << wodi_id << "\n"; cout.flush();
        while (1) {
            int recv_id = wodi_trecvresponse(CORAL_RESP_TAG, &timeout);
            if (recv_id >= 0) {
                assert (recv_id == wodi_id);
                break;
            }
            else {
                wodi_killwork(wodi_id);
                int len = strlen(backup);
                if (pvm_pkint(&len, 1, 1) < 0) {
                    pvm_perror("pvm_pkint");
                    return EOF;
                }

                // send the real msg
                if (pvm_pkbyte((char*)backup, len, 1) < 0) {
                    pvm_perror("pvm_pkbyte");
                    return EOF;
                }

                wodi_sendwork(wodi_id);
                cout << "REDO WODI_SENDWORK with wodi_id = "<< wodi_id << endl;
                cout << "QUERY = " << backup << endl; 
                cout.flush();

            }
        }
#endif
    }
    else { // receive the work from master
        print_cur_time();cout << " about to WODI_RECVWORK\n"; cout.flush();
        wodi_recvwork(-1, CORAL_WORK_TAG);
    }

    int len;

    // before every pvm msg there is an integer indicating the length
    // of the msg
    if (pvm_upkint(&len, 1, 1) < 0) {
        pvm_perror("pvm_upkint");
        return EOF;
    }

    if (pvm_upkbyte(cp, len, 1) < 0) { // assume len <= put_buffer_size
        pvm_perror("pvm_upkbyte");
        return EOF;
    }

    if(type != WODI_MASTER) {
        print_cur_time();
        cout << " WODI_RECVWORK done\n";
        cout << "QUERY = " << cp << endl; 
        cout.flush();
    }
    else {
        time_t cal_time = time(NULL);
        char * s = new char[20];
        strftime(s, 20, "%m/%d %H:%M:%S",localtime(&cal_time));
        (w_out) << s;
        (w_out).flush();
        delete s;
    
        (w_out) << " WODI_RECVRESPONS done with new wodi_id = " << wodi_id << "\n";
        (w_out).flush();
    }

    setg(cp, cp, cp + len); // no putback (at least for now...)

    return(*cp);
}

#include <time.h>

void print_cur_time(void)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    time_t cal_time = time(NULL);
    char * s = new char[20];
    strftime(s, 20, "%m/%d %H:%M:%S",localtime(&cal_time));
    cout << s;
    cout.flush();
    delete[] s;
}
    
#endif // DO_WODI

