// *************************************************************************
//    Tuple.h
// *************************************************************************

#ifndef _Tuple_H_
#define _Tuple_H_

#include "BindEnv.h"
#include "ArgList.h"
#include "Term.h"
#include "Entity.h"
#include "Debug.h"

// *************************************************************************
// A Tuple has an ArgList and some flags.
// We assume that all variables in the tuple are free.  

class Cor_Tuple {

  public:
    Cor_ArgList   *_args;     // array of terms. Length==relation's arity

    short          contid;    // This field and the next are related to
    short          parid;     // return unification optimizations.

    unsigned int   deleted : 1;  // Has the tuple been deleted?
    unsigned int   no_args : 1;  // Set if this tuple doesn't own its _args
    unsigned int   in_tlist : 1; // Is this tuple currently in a TupleList

    unsigned short del_ref_cnt;  // Number of indexes deleted tuple is still
                                 // present in.

    Cor_Tuple() {
        _args = NULL;
        deleted  = 0;
        no_args  = 0;
        in_tlist = 0;
        contid = parid = 0;
    }

    Cor_Tuple(int arity) {
        _args = new Cor_ArgList(arity);
        deleted = 0;
        no_args = 0;
        in_tlist = 0;
        contid = parid = 0;
    }

    Cor_Tuple(Cor_ArgList& a) {
        _args = &a;
        deleted = 0;
        no_args = 0;
        in_tlist = 0;
        contid = parid = 0;
    }

    Cor_Tuple(Cor_ArgList *arglist) {
        _args = arglist;
        deleted = 0;
        no_args = 0;
        in_tlist = 0;
        contid = parid = 0;
    }

   //~Cor_Tuple();

    // True if all arguments are ground.
    int isConstant();

    inline void do_delete()        {
        deleted = 1;
    }

    inline int is_deleted()        { return(deleted); }

    inline int is_in_tlist()        { return(in_tlist); }

    inline void set_no_args(int na=1) { no_args = na; }
    inline int has_no_args()        { return(no_args); }

    inline void set_del_ref(int df) {
        ASSERT(deleted);    // Only set for deleted tuples
        del_ref_cnt = df;
    }

    inline int  decr_del_ref() { return(--del_ref_cnt); }

    inline Cor_ArgList& args() const     { return(*_args); }
    inline int arity() const    { return(_args->count()); }

    // Returns a term that is the i'th argument of
    // the tuple.
    inline Cor_Term *term(int i) const {
        Cor_Term *term = new Cor_Term ((*_args)[i], NULL);
        return term;
    }

    inline Cor_CArg& arg(int i) const { return (*_args)[i]; }

        // Cannot make this inline !! - Praveen
    Cor_CArg& operator [] (int i) const;

    int operator ==(const Cor_Tuple& that) const;

    inline int hash( Cor_BitVector& bv ) const {
        return _args->hash( bv, NULL );
    }

    inline int hash() const { return _args->hash(NULL); }

      // Instantiate all variables that are bound in this bindenv.
    void instantiate( Cor_BindEnv *env );

      // Check for equality
    int equals(Cor_Tuple *that);

    friend ostream& operator<< (ostream& out, const Cor_Tuple& tuple);

    void sprint(char *buff, int *pos) const;
    void printon(ostream *file) const;

    // This handles the reference counting of deleted tuples
    int delete_self(int force_del=0);
};

// *************************************************************************

#endif _Tuple_H_
