// *************************************************************************
//    TupleStack.C
// *************************************************************************

#include <stdlib.h>

#include "TupleStack.h"
#include "Debug.h"

        // The stack can grow, but start small,
        // if we are in a hash-table we won't be allowed to get too large
#define    DEFAULT_SIZE    4

// *************************************************************************
int  Cor_TupleStack::push(HashVal hash, Cor_Tuple *tpl)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (length >= array_size) {

        if (array_size == 0) {
            array_size = DEFAULT_SIZE;
            tuples = new Cor_Tuple* [array_size];
            hashs = new HashVal [array_size];
        } else {

            int len = array_size;
            array_size += array_size;
            Cor_Tuple **tmp_tuples = tuples;
            HashVal *tmp_hashs = hashs;
            tuples = new Cor_Tuple* [array_size];
            hashs = new HashVal [array_size];
            memcpy(tuples, tmp_tuples, len * sizeof(Cor_Tuple*));
            memcpy(hashs, tmp_hashs, len * sizeof(HashVal));
            delete [] tmp_tuples;
            delete [] tmp_hashs;
        }
    }

    hashs[length]    = hash;
    tuples[length++] = tpl;
    //num_items++;

    return(length-1);
}

// *************************************************************************
//  copies contents of stack stk to this one
void Cor_TupleStack::copy(Cor_TupleStack& stk)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    delete [] tuples;
    delete [] hashs;

    length = stk.length;
    array_size = stk.array_size;
    tuples = new Cor_Tuple* [array_size];
    hashs = new HashVal[array_size];

    memcpy(tuples, stk.tuples, length * sizeof(Cor_Tuple*));
    memcpy(hashs, stk.hashs, length * sizeof(HashVal));
}

// *************************************************************************
void Cor_TupleStack::free_all(int deleteTuples)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j;

    if (deleteTuples)
        for (j=0; j < count(); j++) {
            delete tuples[j];
        }

    empty();
}

// *************************************************************************
// *************************************************************************
// Compress this stack if possible, returning the translated
// position of the input position
int  Cor_TupleStack::compress(int pos)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int first=0, last=count()-1, before=0;
    unsigned int num_items = 0;

    cerr << "TupleStack::compress() does not work properly yet.\n";
    ASSERT(0);

    for (int j=0; j < count(); j++)
        if (tuples[j])
            num_items++;

    while (length != num_items) {

        // find the first empty slot
        for ( ; first < count(); first++) {
            if (tuples[first] == NULL)
                break;
        }

        if (first < pos)
            before++;

        // find the last non-empty slot
        for ( ; last >= 0; last--) {
            if (tuples[last] == NULL)
                length--;
            else
                break;
        }

          // Sanity check before swapping
        if (first < last) {
            tuples[first] = tuples[last];
            hashs[first]  = hashs[last];
            length--;
            last--;
        } else {
            TELL;
            ASSERT(0);
        }
    }

    return(pos-before);
}

// *************************************************************************
ostream& operator<< (ostream& out, const Cor_TupleStack& tstk)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j;

    for (j=0; j < (int) tstk.length; j++) {
        if (tstk.tuples[j])
            out << *(tstk.tuples[j]) << endl;
    }

    return(out);
}

// *************************************************************************

#undef    DEFAULT_SIZE

// *************************************************************************
