// *************************************************************************
//    SupOpt.C
// *************************************************************************

#include "SupOpt.h"

#include "SupMagic.h"

#include "Rule.h"
#include "Literal.h"

/*****************************************************************************/

extern int opt_share_functors;

/*****************************************************************************/
/*
    Perform a simple optimization to the supplementary magic rules: replace
    all occurrences of supmagic_r_1 by the right-hand side of the rule defining
    supmagic_r_1 and delete that rule.
*/

void supopt()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;
    int j, nrules;
    Cor_Literal *pred;
    Cor_Literal *newpred;

    Cor_Clause *magicrule;
    extern Cor_Collection *magic_rules;    // Main.C

    nrules = magic_rules->count();

    for (j=0; j < nrules; j++) {
        magicrule = magic_rules->get_ith(j);

        // Delete any rule defining supmagic_r_1.
        if ((magicrule->head->kind == COR_SUPMAGIC)
              && (magicrule->head->pos == 1)) {

            magic_rules->del_ith(j);
            continue;
        } else {

            // Search for and replace uses of supmagic_r_1.
            for (i=0; i < magicrule->num_preds; ++i) {
                pred = magicrule->preds[i];
                if (pred->kind == COR_SUPMAGIC && pred->pos == 1) {
                    // ?Change the following to deal with opt_share_functors.
                    newpred = build_magic_pred(NULL, pred->rule->head,0);
                    magicrule->preds[i] = newpred;
                    newpred->pred = pred->rule->head->pred;
                    newpred->adorn = pred->rule->head->adorn;
                }
            }
        }
    }

      // To compact the array again
      // (only does it if some rules were deleted)
    magic_rules->squash();
}

// *************************************************************************
